"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Overlay = void 0;
const printTree_1 = require("tree-dump/lib/printTree");
const printBinary_1 = require("tree-dump/lib/printBinary");
const util_1 = require("sonic-forest/lib/util");
const util_2 = require("sonic-forest/lib/splay/util");
const OverlayPoint_1 = require("./OverlayPoint");
const MarkerOverlayPoint_1 = require("./MarkerOverlayPoint");
const refs_1 = require("./refs");
const clock_1 = require("../../../json-crdt-patch/clock");
const json_hash_1 = require("../../../json-hash");
const MarkerSlice_1 = require("../slice/MarkerSlice");
const util_3 = require("../../../json-crdt/nodes/rga/util");
/**
 * Overlay is a tree structure that represents all the intersections of slices
 * in the text. It is used to quickly find all the slices that overlap a
 * given point in the text. The overlay is a read-only structure, its state
 * is changed only by calling the `refresh` method, which updates the overlay
 * based on the current state of the text and slices.
 */
class Overlay {
    constructor(txt) {
        this.txt = txt;
        this.root = undefined;
        // ----------------------------------------------------------------- Stateful
        this.hash = 0;
        this.slices = new Map();
    }
    first() {
        return this.root ? (0, util_1.first)(this.root) : undefined;
    }
    iterator() {
        let curr = this.first();
        return () => {
            const ret = curr;
            if (curr)
                curr = (0, util_1.next)(curr);
            return ret;
        };
    }
    splitIterator() {
        let curr = this.first();
        return () => {
            while (curr) {
                const ret = curr;
                if (curr)
                    curr = (0, util_1.next)(curr);
                if (ret instanceof MarkerOverlayPoint_1.MarkerOverlayPoint)
                    return ret;
            }
            return undefined;
        };
    }
    /**
     * Retrieve overlay point or the previous one, measured in spacial dimension.
     */
    getOrNextLower(point) {
        let curr = this.root;
        let result = undefined;
        while (curr) {
            const cmp = curr.cmpSpatial(point);
            if (cmp === 0)
                return curr;
            if (cmp > 0)
                curr = curr.l;
            else {
                const next = curr.r;
                result = curr;
                if (!next)
                    return result;
                curr = next;
            }
        }
        return result;
    }
    find(predicate) {
        let point = this.first();
        while (point) {
            if (predicate(point))
                return point;
            point = (0, util_1.next)(point);
        }
        return undefined;
    }
    refresh(slicesOnly = false) {
        const txt = this.txt;
        let hash = 5381 /* CONST.START_STATE */;
        hash = this.refreshSlices(hash, txt.savedSlices);
        hash = this.refreshSlices(hash, txt.extraSlices);
        hash = this.refreshSlices(hash, txt.localSlices);
        // if (!slicesOnly) this.computeSplitTextHashes();
        return (this.hash = hash);
    }
    refreshSlices(state, slices) {
        const oldSlicesHash = slices.hash;
        const changed = oldSlicesHash !== slices.refresh();
        const sliceSet = this.slices;
        state = (0, json_hash_1.updateNum)(state, slices.hash);
        if (changed) {
            slices.forEach((slice) => {
                let tuple = sliceSet.get(slice);
                if (tuple) {
                    if (slice.isDel && slice.isDel()) {
                        this.delSlice(slice, tuple);
                        return;
                    }
                    const positionMoved = tuple[0].cmp(slice.start) !== 0 || tuple[1].cmp(slice.end) !== 0;
                    if (positionMoved)
                        this.delSlice(slice, tuple);
                    else
                        return;
                }
                tuple = this.insSlice(slice);
                this.slices.set(slice, tuple);
            });
            if (slices.size() < sliceSet.size) {
                sliceSet.forEach((tuple, slice) => {
                    const mutSlice = slice;
                    if (mutSlice.isDel) {
                        if (!mutSlice.isDel())
                            return;
                        this.delSlice(slice, tuple);
                    }
                });
            }
        }
        return state;
    }
    point(id, anchor) {
        return new OverlayPoint_1.OverlayPoint(this.txt.str, id, anchor);
    }
    mPoint(marker, anchor) {
        return new MarkerOverlayPoint_1.MarkerOverlayPoint(this.txt.str, marker.start.id, anchor, marker);
    }
    /**
     * Retrieve an existing {@link OverlayPoint} or create a new one, inserted
     * in the tree, sorted by spatial dimension.
     */
    upsertPoint(point) {
        const newPoint = this.point(point.id, point.anchor);
        const pivot = this.insPoint(newPoint);
        if (pivot)
            return [pivot, false];
        return [newPoint, true];
    }
    /**
     * Inserts a point into the tree, sorted by spatial dimension.
     * @param point Point to insert.
     * @returns Returns the existing point if it was already in the tree.
     */
    insPoint(point) {
        let pivot = this.getOrNextLower(point);
        if (!pivot)
            pivot = (0, util_1.first)(this.root);
        if (!pivot) {
            this.root = point;
            return;
        }
        else {
            if (pivot.cmp(point) === 0)
                return pivot;
            const cmp = pivot.cmpSpatial(point);
            if (cmp < 0)
                (0, util_1.insertRight)(point, pivot);
            else
                (0, util_1.insertLeft)(point, pivot);
        }
        if (this.root !== point)
            this.root = (0, util_2.splay)(this.root, point, 10);
        return undefined;
    }
    delPoint(point) {
        this.root = (0, util_1.remove)(this.root, point);
    }
    insMarker(slice) {
        const point = this.mPoint(slice, 0 /* Anchor.Before */);
        const pivot = this.insPoint(point);
        if (!pivot) {
            point.refs.push(slice);
            const prevPoint = (0, util_1.prev)(point);
            if (prevPoint)
                point.layers.push(...prevPoint.layers);
        }
        return [point, point];
    }
    insSlice(slice) {
        if (slice instanceof MarkerSlice_1.MarkerSlice)
            return this.insMarker(slice);
        const txt = this.txt;
        const str = txt.str;
        let startPoint = slice.start;
        let endPoint = slice.end;
        const startIsStringRoot = (0, clock_1.equal)(startPoint.id, str.id);
        if (startIsStringRoot) {
            const firstVisibleChunk = (0, util_3.firstVis)(txt.str);
            if (firstVisibleChunk) {
                startPoint = txt.point(firstVisibleChunk.id, 0 /* Anchor.Before */);
                const endIsStringRoot = (0, clock_1.equal)(endPoint.id, str.id);
                if (endIsStringRoot) {
                    endPoint = txt.point(firstVisibleChunk.id, 0 /* Anchor.Before */);
                }
            }
        }
        const [start, isStartNew] = this.upsertPoint(startPoint);
        const [end, isEndNew] = this.upsertPoint(endPoint);
        start.refs.push(new refs_1.OverlayRefSliceStart(slice));
        end.refs.push(new refs_1.OverlayRefSliceEnd(slice));
        if (isStartNew) {
            const beforeStartPoint = (0, util_1.prev)(start);
            if (beforeStartPoint)
                start.layers.push(...beforeStartPoint.layers);
        }
        if (isEndNew) {
            const beforeEndPoint = (0, util_1.prev)(end);
            if (beforeEndPoint)
                end.layers.push(...beforeEndPoint.layers);
        }
        const isCollapsed = startPoint.cmp(endPoint) === 0;
        let curr = start;
        while (curr !== end && curr) {
            curr.addLayer(slice);
            curr = (0, util_1.next)(curr);
        }
        if (!isCollapsed) {
        }
        else {
            start.addMarker(slice);
        }
        return [start, end];
    }
    delSlice(slice, [start, end]) {
        this.slices.delete(slice);
        let curr = start;
        do {
            curr.removeLayer(slice);
            curr.removeMarker(slice);
            curr = (0, util_1.next)(curr);
        } while (curr && curr !== end);
        start.removeRef(slice);
        end.removeRef(slice);
        if (!start.refs.length)
            this.delPoint(start);
        if (!end.refs.length && start !== end)
            this.delPoint(end);
    }
    // ---------------------------------------------------------------- Printable
    toString(tab = '') {
        const printPoint = (tab, point) => {
            return (point.toString(tab) +
                (0, printBinary_1.printBinary)(tab, [
                    !point.l ? null : (tab) => printPoint(tab, point.l),
                    !point.r ? null : (tab) => printPoint(tab, point.r),
                ]));
        };
        return (`${this.constructor.name} #${this.hash.toString(36)}` +
            (0, printTree_1.printTree)(tab, [!this.root ? null : (tab) => printPoint(tab, this.root)]));
    }
}
exports.Overlay = Overlay;
